@@ -54,7 +54,7 @@ OSHELL_NAME = "osh"
 HTML_ESCAPES = ["\x1b[?1155;", "h",
                 "\x1b[?1155l"]
 
-USER_RE    = re.compile(r"^[a-z][a-z0-9\-]*$")                # Allowed user names
+USER_RE    = re.compile(r"^[a-z][a-z0-9]*$")                  # Allowed user names
 HOST_RE    = re.compile(r"^[a-z][a-z0-9\-\*\?\[\]]*$")        # Allowed host names
 SESSION_RE = re.compile(r"^[a-zA-Z\*\?\[\]][\w\*\?\[\]]*$")   # Allowed session names
 
@@ -271,10 +271,10 @@ class TerminalClient(packetserver.RPCLink, packetserver.PacketClient):
             Widget_server.listen(self.widget_port, address="localhost")
             logging.warning("GraphTerm widgets listening on %s:%s", "localhost", self.widget_port)
 
-    def remote_response(self, term_name, websocket_id, message_list):
-        self.send_request_threadsafe("response", term_name, websocket_id, message_list)
+    def remote_response(self, term_name, websocket_id, message_list, _content=None):
+        self.send_request_threadsafe("response", term_name, websocket_id, message_list, _content=_content)
 
-    def remote_request(self, term_name, from_user, req_list):
+    def remote_request(self, term_name, from_user, req_list, _content=None):
         """
         Setup commands:
           reconnect <response_id> <host_settings>
@@ -286,10 +286,12 @@ class TerminalClient(packetserver.RPCLink, packetserver.PacketClient):
           click_paste <text> <file_url> {command:, clear_last:, normalize:, enter:}
           paste_command <text>
           get_finder <kind> <directory>
-          save_data <save_params> <filedata>
-          open_notebook <filepath> <prompts> <content>
+          save_data <save_params> <filedata>|None
+          open_notebook <filepath> <share> <prompts> <content>
           close_notebook <discard>
           save_notebook <filepath> <input_data> <params>
+          submit_notebook <source_terminal> <filepath> <file_data>
+          note_lock <offset>
           add_cell <new_cell_type> <init_text> <before_cell_index>
           select_cell <cell_index> <move_up> <next_code>
           select_page <move_up> <endpoint> <slide>
@@ -311,21 +313,22 @@ class TerminalClient(packetserver.RPCLink, packetserver.PacketClient):
         try:
             lterm_cookie, blobs = self.terms.get(term_name, [None, None])
             resp_list = []
-            for cmd in req_list:
+            for j, cmd in enumerate(req_list):
                 action = cmd.pop(0)
                 if action == "shutdown":
                     if cmd:
                         logging.warning("SHUTDOWN %s", cmd[0])
                     self.shutdown()
+                    IO_loop.add_callback(host_shutdown)
 
                 elif action == "reconnect":
                     self.term_reconnect(cmd[1])
                     if self.lineterm:
                         response_id = cmd[0]
                         self.lineterm.reconnect(term_name, response_id)
-                        widget_pagelet = WidgetStream.get_widget_pagelet(lterm_cookie)
-                        if widget_pagelet:
-                            self.send_request_threadsafe("response", term_name, response_id, [["terminal", "graphterm_widget", widget_pagelet]])
+                        widget_pagelet_args = WidgetStream.get_pagelet_args(lterm_cookie)
+                        if widget_pagelet_args:
+                            self.send_request_threadsafe("response", term_name, response_id, [["terminal", "graphterm_widget", widget_pagelet_args]])
 
                 elif action == "set_email":
                     gterm.write_email(cmd[0])
@@ -357,9 +360,14 @@ class TerminalClient(packetserver.RPCLink, packetserver.PacketClient):
                         chat_widget.send_packet(msg.encode(self.host_settings["term_encoding"], "ignore"))
 
                 elif action == "save_data":
-                    # save_data <save_params> <filedata>
+                    # save_data <save_params> <filedata>|None
                     if self.lineterm:
-                        self.lineterm.save_data(term_name, cmd[0], cmd[1])
+                        if cmd[1] is None:
+                            assert j == len(req_list)-1, "save_data with content must be the last request"
+                            data = _content
+                        else:
+                            data = cmd[1]
+                        self.lineterm.save_data(term_name, cmd[0], data)
 
                 elif action == "save_prefs":
                     # save_prefs <prefs_dict>
@@ -373,9 +381,9 @@ class TerminalClient(packetserver.RPCLink, packetserver.PacketClient):
                         self.paste_command(term_name, paste_text)
 
                 elif action == "open_notebook":
-                    # open_notebook <filepath> <prompts> <content>
+                    # open_notebook <filepath> <share> <prompts> <content>
                     if self.lineterm:
-                        self.lineterm.open_notebook(term_name, cmd[0], cmd[1], cmd[2])
+                        self.lineterm.open_notebook(term_name, cmd[0], cmd[1], cmd[2], cmd[3])
 
                 elif action == "close_notebook":
                     # close_notebook <discard>
@@ -387,6 +395,23 @@ class TerminalClient(packetserver.RPCLink, packetserver.PacketClient):
                     if self.lineterm:
                         self.lineterm.save_notebook(term_name, cmd[0], cmd[1], cmd[2])
 
+                elif action == "submit_notebook":
+                    # submit_notebook <source_terminal> <filepath> <filedata>
+                    try:
+                        with open(cmd[1], "w") as f:
+                            f.write(cmd[2])
+                        status_msg = "Submitted notebook %s" % cmd[1]
+                    except Exception, excp:
+                        status_msg = "Error in submitting notebook %s: %s" % (cmd[1], excp)
+                        logging.error("TerminalClient.remote_request: %s", status_msg)
+
+                    self.remote_response(term_name, "", [["terminal", "remote_alert", [cmd[0], status_msg] ]])
+
+                elif action == "note_lock":
+                    # note_lock <offset>
+                    if self.lineterm:
+                        self.lineterm.note_lock(term_name, cmd[0])
+
                 elif action == "add_cell":
                     # add_cell <new_cell_type> <init_text> <before_cell_index>
                     if self.lineterm:
@@ -577,13 +602,18 @@ class TerminalClient(packetserver.RPCLink, packetserver.PacketClient):
                                       dict(status=status, last_modified=last_modified,
                                            etag=etag,
                                            content_type=content_type, content_length=content_length,
-                                           content_b64=content_b64)])
+                                           content_b64=None if content_b64 else "")])
+                    if content_b64:
+                        # Send response with file content right away
+                        self.remote_response(term_name, "", resp_list, _content=content_b64)
+                        resp_list = []
 
                 elif action == "errmsg":
                     logging.warning("remote_request: ERROR %s", cmd[0])
                 else:
                     raise Exception("Invalid action: "+action)
-            self.remote_response(term_name, "", resp_list);
+            if resp_list:
+                self.remote_response(term_name, "", resp_list)
         except Exception, excp:
             if isinstance(excp, gterm.MsgException):
                 errmsg = str(excp)
@@ -630,8 +660,8 @@ else:
 
 class WidgetStream(object):
     _all_widgets = []
-    _term_pagelets = {}
     _chat_widgets = {}
+    _pagelet_widgets = {}
 
     startseq = "\x1b[?1155;"
     startdelim = "h"
@@ -643,6 +673,7 @@ class WidgetStream(object):
         self.packet_cookie = ""
         self.chat_cookie = ""
         self.widget_token = ""
+        self.last_pagelet_args = None
         self._all_widgets.append(self)
         self.stream.set_close_callback(self.on_close)
 
@@ -652,9 +683,18 @@ class WidgetStream(object):
         return cls._chat_widgets.get(lterm_cookie)
 
     @classmethod
-    def get_widget_pagelet(cls, lterm_cookie):
-        """Return widget pagelet associated with terminal"""
-        return cls._term_pagelets.get(lterm_cookie)
+    def get_pagelet_widget(cls, lterm_cookie):
+        """Return pagelet WidgetStream instance"""
+        return cls._pagelet_widgets.get(lterm_cookie)
+
+    @classmethod
+    def get_pagelet_args(cls, lterm_cookie):
+        """Return widget pagelet args associated with terminal"""
+        pwidget = cls._pagelet_widgets.get(lterm_cookie)
+        if pwidget:
+            return pwidget.last_pagelet_args
+        else:
+            return None
 
     @classmethod
     def shutdown_all(cls):
@@ -664,9 +704,19 @@ class WidgetStream(object):
     def on_close(self):
         try:
             if self.chat_cookie:
+                # Chat widget
                 self.set_chat_status(False)
                 self._chat_widgets.pop(self.chat_cookie, None)
                 self.chat_cookie = ""
+
+            if self.last_pagelet_args:
+                # Pagelet widget
+                headers = {"content_type": "text/html"}
+                headers["x_gterm_response"] = "pagelet"
+                headers["x_gterm_parameters"] = {}
+                # Send blank pagelet to erase widget
+                self.send_widget_content({"validated": True, "headers": headers}, "")
+                self._pagelet_widgets.pop(self.packet_cookie, None)
             self._all_widgets.remove(self)
             self.stream = None
         except Exception:
@@ -711,7 +761,6 @@ class WidgetStream(object):
         term_info = TerminalClient.all_cookies.get(self.packet_cookie)
         if not term_info:
             return self.shutdown()
-
         host_connection, term_name = term_info
 
         headers, content = lineterm.parse_headers(data)
@@ -744,12 +793,25 @@ class WidgetStream(object):
         else:
             params = {"validated": True, "headers": headers}
             args = [params, base64.b64encode(content) if content else ""]
-            if resp_type == "pagelet":
-                self._term_pagelets[self.packet_cookie] = args
-            host_connection.send_request_threadsafe("response", term_name, "", [["terminal", "graphterm_widget", args]])
+            if not resp_type or resp_type == "pagelet":
+                # HTML pagelet widget
+                prev_widget = self.get_pagelet_widget(self.packet_cookie)
+                if prev_widget and prev_widget is not self:
+                    # Only one active pagelet widget per terminal at a time
+                    prev_widget.shutdown()
+                self._pagelet_widgets[self.packet_cookie] = self
+                self.last_pagelet_args = args
+            self.send_widget_content(*args)
+
         self.packet_cookie = ""
         self.next_packet()
 
+    def send_widget_content(self, params, content):
+        term_info = TerminalClient.all_cookies.get(self.packet_cookie)
+        if term_info:
+            host_connection, term_name = term_info
+            host_connection.send_request_threadsafe("response", term_name, "", [["terminal", "graphterm_widget", [params, content] ]])
+
     def send_packet(self, data):
         if self.stream:
             self.stream.write(data)
@@ -835,6 +897,11 @@ def gterm_connect(host_name, server_addr, server_port=gterm.DEFAULT_HOST_PORT, c
 
     return (host_connection, host_secret, trace_shell)
 
+def host_shutdown():
+    global Gterm_host
+    gterm_shutdown(Trace_shell)
+    Gterm_host = None
+
 def run_host(options, args):
     global Gterm_host, Host_secret, Trace_shell, Xterm, Killterm
     host_name = args[0]
@@ -874,11 +941,6 @@ def run_host(options, args):
     Xterm = Gterm_host.xterm
     Killterm = Gterm_host.remove_term
 
-    def host_shutdown():
-        global Gterm_host
-        gterm_shutdown(Trace_shell)
-        Gterm_host = None
-
     def sigterm(signal, frame):
         logging.warning("SIGTERM signal received")
         IO_loop.add_callback(host_shutdown)
@@ -897,7 +959,8 @@ def run_host(options, args):
         print >> sys.stderr, "Interrupted"
 
     finally:
-        IO_loop.add_callback(host_shutdown)
+        if IO_loop:
+            IO_loop.add_callback(host_shutdown)
 
 
 def main():
